
//	Tor Tray 0.9.6
//	Simple tray loader and controller for Tor
//
//	Copyright (C) 2005 Ferruh Mavituna
//	http://ferruh.mavituna.com, ferruh@mavituna.com
//	========================================================================================
//
//	Tor,
//	http://tor.eff.org/
//
//	Thanks;
//	- Some code snippets based on Timothy Legge's (tlegge@rogers.com) Printers 1.0 (GPL)
//	- LIBCTINY.LIB by Matt Pietrek - http://msdn.microsoft.com/msdnmag/issues/01/01/hood/
//		Source Code;
//		http://download.microsoft.com/download/2/8/c/28c4ace3-f5ed-4e14-bc64-3d563b807dfb/Hood0101.exe
// 
//	========================================================================================
//
//	This program is free software; you can redistribute it and/or
//	modify it under the terms of the GNU General Public License
//	as published by the Free Software Foundation; either version 2
//	of the License, or (at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


/* 
HISTORY
========================================================================================
01.01.05
	- New

02.21.05
	- Fixes
	- Showmessage Feature

10.01.05 (0.9.6)
	- LIBCTINY.LIB added (reduce output size)
	- Fixed some small bugs (start check etc.)
	
TODO;
	- Remember last show message
	- Add Launch privoxy option
	- Add an config file for advanced configuration
		- Tor Path
		- Privoxy Path
		- Launch command (parameter support)
		
*/ 

#include "tortray.h"
#pragma comment(linker, "/OPT:NOWIN98")

// Switch Menu Text (Start/Stop Tor)
VOID switchMenu(BOOL bStart);

// Launch Tor
DWORD WINAPI launchTor(LPVOID lpParam);

// Sleep & Hide Baloon Hack
DWORD WINAPI hideBalloon(LPVOID lpParam);

// Report, Handle Errors
VOID err(LPTSTR lpszFunction);

// Start/Stop Tor
VOID startCommand();

// Read from console
VOID readFromPipe(); 

// Run Tor Thread
BOOL createChildProcess(); 

// Terminate Tor
VOID closeTor();

VOID switchBaloon();

// Show NotifyIcon Balloon
VOID showBalloon(LPCSTR szText, LPCSTR szTitle, DWORD dwFlag);

// Change NotifyIcon Tooltip
VOID setTooltip(LPSTR lpszTip);

// Tor clean Start Check
BOOL startCheck();

// Struct for menu item data
typedef struct tagMYITEM 
{ 
    HFONT hfont; 
    int   cchItemText; 
    char  szItemText[1]; 
} MYITEM; 

// Globals
HINSTANCE ghThisInst;			// Global value for this instance of the Application
HICON hIcon;					// Handle to the Tray Icon
NOTIFYICONDATA tnd;				// Structure for the Tray Icon Data
PROCESS_INFORMATION piProcInfo; // Process Information (tor console)

// Global ThreadId for hide hack
HANDLE hndHide, hChildStdoutRd, hChildStdoutWr, hChildStdoutRdDup; 
BOOL m_bWin2K, torActive, bBalloon;
HMENU hMenu;	// Handle to the pop-up menu

// Menu & Icon Related
BOOL AddTaskBarIcon(HWND hwnd, UINT uID, LPSTR lpszTip) ;
LRESULT CALLBACK WindowFunc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
DWORD ShowMenu(HWND hwnd, HMENU hMenu, POINT pt);


// Entry Point
int WINAPI WinMain(HINSTANCE hThisInst, HINSTANCE hPrevInst, LPSTR lpszArgs, int nWinMode){
	bBalloon = TRUE;

	// Check for Shell showballon support
	#ifdef SYSTEMTRAY_USEW2K
		OSVERSIONINFO os = { sizeof(os) };
		GetVersionEx(&os);
		m_bWin2K = ( VER_PLATFORM_WIN32_NT == os.dwPlatformId && os.dwMajorVersion >= 5 );
	#else
		m_bWin2K = FALSE;
	#endif

	MSG msg;					// Message to be used in main message loop
	HWND hwnd;					// Handle to the Window
	WNDCLASSEX wcl;				// Structure for the Window class information
	ghThisInst = hThisInst;		// Set global value for this instance of the Application
	
	//	Set the information for the Window Class Structure
	wcl.cbSize   = sizeof(WNDCLASSEX);
	wcl.style = CS_HREDRAW | CS_VREDRAW ;
	wcl.lpfnWndProc = WindowFunc;		// Name of Call Back Window Function
	wcl.cbClsExtra = 0;					
	wcl.cbWndExtra = 0;
	wcl.hInstance = hThisInst;			// Instance Handle
	wcl.hIcon = NULL;					// No Icon for the Window
	wcl.hCursor = NULL;					// No Cursor for the Window
	wcl.hbrBackground = NULL;	
	wcl.lpszMenuName = NULL;			// No Main Menu	
	wcl.lpszClassName = szClassName;	// Class Name
	wcl.hIconSm = NULL;

	if(!RegisterClassEx (&wcl)) return 0;	// Register Window Class
	
	// Create the window, but do not display it.
	hwnd = CreateWindow(szClassName, "", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hThisInst, NULL);
	
	// Load the Icon and get the handle
	hIcon = LoadIcon(hThisInst, (const char *) ID_TASKICON);	
	
	//Add Notify Icon
	AddTaskBarIcon(hwnd, IDC_UPDATE, _T("Tor Tray")); 

	// Launch Tor
	#ifdef STARTONDEFAULT
		startCommand();
	#endif

	
	//Enter main message loop for the application
	while(GetMessage (&msg, NULL, 0, 0)){
		TranslateMessage(&msg);
		DispatchMessage(&msg);	
	}
	
	return msg.wParam;
}

// Clean Start Check
BOOL startCheck()
{
	if( OpenProcess(PROCESS_ALL_ACCESS,FALSE,piProcInfo.dwProcessId)==NULL ){
		showBalloon("Tor is closed or Access Problem", "Tor Access Error", NIIF_ERROR);
		switchMenu(FALSE);
		return FALSE;

	}else{
		return TRUE;
		switchMenu(TRUE);
	}
	
}

// Execute and Control Tor
DWORD WINAPI launchTor(LPVOID lpParam){
	// Execute Tor
	showBalloon("Tor is starting ...", "Tor Info", NIIF_INFO);
	setTooltip(TOOL_STARTTOR);

	// Set the bInheritHandle flag so pipe handles are inherited. 
	SECURITY_ATTRIBUTES saAttr; 
	BOOL hndSuccess; 
	saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
	saAttr.bInheritHandle = TRUE; 
	saAttr.lpSecurityDescriptor = NULL; 

	// Create a pipe for the child process's STDOUT. 
    if (! CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0))
			err("Child pipe failed");
      
	// Create noninheritable read handle and close the inheritable read handle. 
    hndSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdoutRd, GetCurrentProcess(), &hChildStdoutRdDup , 0, FALSE, DUPLICATE_SAME_ACCESS);
    
	switchMenu(TRUE);

	if(!hndSuccess)
		err("DuplicateHandle failed");

	CloseHandle(hChildStdoutRd);

	// Now create the child process. 
	if(!createChildProcess()){
		showBalloon("Tor.exe couldn't launch or find !", "Tor Launch Error !", NIIF_ERROR);
		return 1;
	}

	showBalloon("Tor Started", "Tor Info", NIIF_INFO);

	// Sleep 2 sc. for check TOR clean Startup
	Sleep(2000); 
	if(!startCheck()){
		return 1;
		switchMenu(TRUE);
	}

	
	readFromPipe(); 

	return 0;
}

LRESULT CALLBACK WindowFunc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam){	
	POINT point;	// Point to hold co-ordinate to display menu
	int result = 0;  // Result of calling ModifyINI function

	switch (message){
	case WM_COMMAND:	// Received when pop-up menu clicked
		
		switch(LOWORD(wParam)){	// Menu Item ID
			case IDM_EXIT:	// When Exit Menu Item Clicked
				
				// Close Tor if it's active
				closeTor();

				// If icon exists destroy it
				if(hIcon){
					Shell_NotifyIcon(NIM_DELETE, &tnd);
					DestroyMenu(hMenu);
					DestroyIcon(hIcon);
				}

				// Put end program message on Queue
				PostQuitMessage(0);
				
				return 0;
				
			//case IDM_ABOUT:
			//	FIX: Show About Box !
			//	return 0;
		
			case IDM_SHOWMESSAGE: // Disable Flag for showBallon Messages 
				switchBaloon();
				
				return 0;
				

			case IDM_STARTTOR:
				// Start on Double click but do not stop!
				if(!torActive)
					startCommand();
				
				else
					closeTor();
							
				return 0;

			case IDM_ABOUT:
				ShellExecute(NULL, "open", "http://ferruh.mavituna.com/tortray/",NULL, NULL, SW_SHOWNORMAL); 
				
		}
		break;
		
		// Message received when the window is created
		case WM_CREATE:
			// Load Menu
			hMenu = LoadMenu(ghThisInst, "MENU"); // Load the Pop-up Menu
			hMenu = GetSubMenu(hMenu, 0);

			// Set Default menu
			SetMenuDefaultItem(hMenu, IDM_STARTTOR, FALSE);

			break;
			
		//	Message received when Tray Icon gets focus
		case MYWM_NOTIFYICON:
			switch (lParam){
				//	Display the Menu When Right or left Button clicked on the Icon
				case WM_RBUTTONDOWN:	
					GetCursorPos(&point);
					result = ShowMenu(hwnd, hMenu, point);
					return 0;
					
				case WM_LBUTTONDOWN:
					GetCursorPos(&point);
					result = ShowMenu(hwnd, hMenu, point);
					return 0;

				case WM_LBUTTONDBLCLK:
					startCommand();
					return 0;
			}
							
			return 0;

		case WM_DESTROY:
			// If icon exists destroy it
			if(hIcon){
				Shell_NotifyIcon(NIM_DELETE, &tnd);
				DestroyMenu(hMenu);
				DestroyIcon(hIcon);
			}
			break;

		case WM_QUIT:
			// If icon exists destroy it
			if(hIcon){
				Shell_NotifyIcon(NIM_DELETE, &tnd);
				DestroyMenu(hMenu);
				DestroyIcon(hIcon);
			}
			break;

	}
	
	return DefWindowProc(hwnd, message, wParam, lParam);
}


// Sends Start Command (if active Close otherwise run Tor)
VOID startCommand(){
		HANDLE torThread;

		// Instead of readfileEx creata new thread otherwise we can not control output realtime
		if(!torActive){
            torThread = CreateThread(NULL, 0, launchTor, NULL, 0, NULL);		
				if(torThread == NULL)
					err("CreateThread failed");

				switchMenu(TRUE);
					
				CloseHandle(torThread);
		}else{
			closeTor();
		}
}

// AddTaskBarIcon
BOOL AddTaskBarIcon(HWND hwnd, UINT uID, LPSTR lpszTip){ 
    BOOL res; 
	
    tnd.cbSize = sizeof(NOTIFYICONDATA); 
    tnd.hWnd = hwnd; 
    tnd.uID = uID; 
    tnd.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP; 
    tnd.uCallbackMessage = MYWM_NOTIFYICON; 
    tnd.hIcon = hIcon; 

	lstrcpyn(tnd.szTip, _T("Tor Tray"), sizeof(tnd.szTip)); 
	res = Shell_NotifyIcon(NIM_ADD, &tnd); 
	
    if (hIcon) 
        DestroyIcon(hIcon); 
	
    return res; 
} 


// Set Tray Tooltip
VOID setTooltip(LPSTR lpszTip){
	tnd.uFlags = NIF_TIP; 
	lstrcpyn(tnd.szTip, lpszTip, sizeof(tnd.szTip)); 
	Shell_NotifyIcon(NIM_MODIFY, &tnd);
}

// Show Tray Balloon
VOID showBalloon(LPCSTR szText, LPCSTR szTitle, DWORD dwFlag){ 
	// Don't show Balloon if no support
	if(!m_bWin2K || !bBalloon)return;

	//Terminate old one
	if(hndHide != NULL)
		TerminateThread(hndHide, 0);

	tnd.uFlags = NIF_INFO;
	_tcsncpy(tnd.szInfo, szText, 256);			// Balloon Text
    _tcsncpy(tnd.szInfoTitle, szTitle, 64);		// Balloon Title
	tnd.dwInfoFlags = dwFlag;					//NIIF_NONE, NIIF_INFO, NIIF_WARNING, NIIF_ERROR
	
	if(! Shell_NotifyIcon(NIM_MODIFY, &tnd))
		err("Shell NotifyIcon");

	//Clear
	tnd.szInfoTitle[0] = _T('\0');
    tnd.szInfo[0] = _T('\0');

	hndHide = CreateThread(NULL, 0, hideBalloon, NULL, 0, NULL);
		if(hndHide == NULL)
			err("CreateThread | hndHide");
	
	CloseHandle(hndHide);

	return;
}

// Hide balloon (NotifyIcon timeout hack !), Any idea to fix with utimeout?
DWORD WINAPI hideBalloon(LPVOID lpParam){

	Sleep(BALLOONTIMEOUT);
	
	tnd.uFlags = NIF_INFO;
	Shell_NotifyIcon(NIM_MODIFY, &tnd);

	return 0;
}

// Show Menu
DWORD ShowMenu(HWND hwnd, HMENU hMenu, POINT pt){
	int  x =0;
	DrawMenuBar(hwnd); 
	SetForegroundWindow(hwnd);
	
	/* Display the pop-up menu */
	TrackPopupMenu(
		hMenu,	// Menu handle
		#if (WINVER >= 0x400)
				TPM_TOPALIGN |
		#endif
		TPM_LEFTALIGN | TPM_RIGHTBUTTON,	// Flags as per docs
		pt.x,	// X Co-ordinate
		pt.y, //+ScreenY,
		0, 
		hwnd,	// Parent Window
		NULL);				
	
	PostMessage(hwnd, WM_NULL, 0, 0);
	return 0;
}

// Read output
VOID readFromPipe(){ 
   DWORD dwRead; //, ballIcon = NIIF_INFO;
   CHAR chBuf[BUFSIZE] = {0}; 

	// Read output from the Tor
   for (;;){ 
		if( !ReadFile( hChildStdoutRdDup, chBuf, BUFSIZE, &dwRead, NULL) || dwRead == 0) break; 
		
		// Warning
		if( (strstr(chBuf, "warn") != NULL) | (strstr(chBuf, "error") != NULL) ){
			//ballIcon = NIIF_WARNING;
		    showBalloon(chBuf, "Tor Warning", NIIF_WARNING);
		}
		
		// Info
		else if( (strstr(chBuf, "info") != NULL) | (strstr(chBuf, "notice") != NULL) ){
			//ballIcon = NIIF_INFO;
			
			if(strstr(chBuf, "working") != NULL)
				showBalloon(chBuf, "Tor is Working!", NIIF_INFO);

			else
				showBalloon(chBuf, "Tor Info", NIIF_INFO);
		}
	}
} 

// Load Tor
BOOL createChildProcess(){ 
   STARTUPINFO siStartInfo;
   BOOL bFuncRetn = FALSE; 
 
// Set up members of the PROCESS_INFORMATION structure. 
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );
 
// Set up members of the STARTUPINFO structure. 
   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO); 
   siStartInfo.hStdError = hChildStdoutWr;
   siStartInfo.hStdOutput = hChildStdoutWr;
   siStartInfo.dwFlags |= STARTF_USESTDHANDLES;
   siStartInfo.dwFlags |= STARTF_USESHOWWINDOW;
   siStartInfo.wShowWindow = SW_HIDE; // SW_SHOW
 
// Create the child process. 
    bFuncRetn = CreateProcess(NULL, 
	  TORPATH,			// command line 
      NULL,				// process security attributes 
      NULL,				// primary thread security attributes 
      TRUE,				// handles are inherited 
      0,				// creation flags 
      NULL,				// use parent's environment 
      NULL,				// use parent's current directory 
      &siStartInfo,		// STARTUPINFO pointer 
      &piProcInfo);		// receives PROCESS_INFORMATION 
   
	if (bFuncRetn == 0) {
		err("Tor not found or Launch failed ");
		return FALSE;		
	}

   else{
      CloseHandle(piProcInfo.hProcess);
      CloseHandle(piProcInfo.hThread);
   }

   return TRUE;
}


// Close Tor
VOID closeTor(){
	showBalloon("Tor is closing...", "Tor Info", NIIF_INFO);

	// Get handle again because we can not get HANDLEs directly on console apps
	HANDLE pHandle = OpenProcess(PROCESS_ALL_ACCESS,FALSE,piProcInfo.dwProcessId);
	TerminateProcess(pHandle, 0);
    CloseHandle(pHandle);

	showBalloon("Sucessfully closed!", "Tor Info", NIIF_INFO);

	switchMenu(FALSE);
	setTooltip(TOOL_STOPTOR);
}

//Report Error 
VOID err(LPTSTR lpszFunction){ 
    TCHAR szBuf[255]; 
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError(); 

    FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL);

    wsprintf(szBuf, "%s failed with error %d: %s\nPut TorTray.exe to the same folder with tor.exe", lpszFunction, dw, lpMsgBuf);  
	
	if(m_bWin2K)
		showBalloon(szBuf,"Error Details",NIIF_ERROR);
	
	else
		MessageBox(NULL, szBuf, "Critical Error", MB_OK|MB_ICONEXCLAMATION|MB_DEFBUTTON1); 

    LocalFree(lpMsgBuf);
}

// Menu Item
VOID switchMenu(BOOL bStart){
			MENUITEMINFO mItem;
			MYITEM *pMyItem;
			torActive = bStart;

			// Edit Menu
			mItem.cbSize = sizeof(MENUITEMINFO);
			mItem.fMask = MIIM_STRING ;

	        pMyItem = (MYITEM *) LocalAlloc(LMEM_FIXED, sizeof(MYITEM) + BUFSIZE); 
			
			mItem.dwTypeData = pMyItem->szItemText; 
			mItem.cch = BUFSIZE; 
			GetMenuItemInfo(hMenu, IDM_STARTTOR, FALSE, &mItem); 
			pMyItem->cchItemText = mItem.cch; 

			// Change Text
			if(!bStart)
				lstrcpyn(pMyItem->szItemText, MENU_STARTTOR, BUFSIZE); 

            else
				lstrcpyn(pMyItem->szItemText, MENU_STOPTOR, BUFSIZE); 


			// Set Menu
			mItem.dwTypeData = pMyItem->szItemText; 
			SetMenuItemInfo(hMenu, IDM_STARTTOR, FALSE, &mItem);

}

// Switch showmessage
VOID switchBaloon()
{
	if(bBalloon)
		CheckMenuItem(hMenu, IDM_SHOWMESSAGE, MF_BYCOMMAND | MF_UNCHECKED); 

	else
		CheckMenuItem(hMenu, IDM_SHOWMESSAGE, MF_BYCOMMAND | MF_CHECKED); 


	bBalloon = !bBalloon;

}
